#include "STM32F4.h"

#include "tjpgd.h"
#include "sdio_sd.h"

#include "OLED.inc"

#include "PC_FAT.h"

#define DR_FLAC_IMPLEMENTATION
#include "dr_flac.h"

#define W 160
#define H 128

#define L 2000 /* 24kHz/12Hz */

#define CPU_FREQ 168000000
#define F 24000 

#define TIM_PERIOD ((u32)((CPU_FREQ/8)/F))

#define DAC_DHR12R1 0x40007408

#define PAUSE GPIO_ReadInputDataBit(GPIOA,GPIO_Pin_0)

#define SC (*(volatile u32*)0x1000FFF8)
#define SN (*(volatile u32*)0x1000FFFC)

typedef struct
{
 const BYTE* jpic;
 WORD jsize;
 WORD joffset;
}
IODEV;

IODEV devid;
JDEC jdec;

u8 WORK[3100];
u8 JPEGBuffer[32*1024];

#pragma pack(16)
u16 VideoBuffer[W*H];

#pragma pack(16)
u16 AudioBuffer[L<<1];

FIL *fv;

char SS[10];

UINT in_func(JDEC *jd,BYTE *buff,UINT nbyte)
{
 IODEV *dev=(IODEV*)jd->device;
 nbyte=dev->jsize-dev->joffset>nbyte?nbyte:dev->jsize-dev->joffset;
 if(buff)memcpy(buff,dev->jpic+dev->joffset,nbyte);
 dev->joffset+=nbyte;
 return nbyte;
}

UINT out_func(JDEC *jd, void *bitmap,JRECT *rect)
{
 register WORD *src=(WORD*)bitmap,*dst=VideoBuffer+(rect->top*W+rect->left);
 register UINT y=rect->bottom-rect->top+1,bws=rect->right-rect->left+1;
 while(y--)
 {
  memcpy(dst,src,bws<<1);
  src+=bws;
  dst+=W;
 }
 return 1;
}

void FSMC_GPIO_Init(void)
{
 GPIO_InitTypeDef gpio;
 
 RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOB|RCC_AHB1Periph_GPIOD|RCC_AHB1Periph_GPIOE,ENABLE);
 
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource14,GPIO_AF_FSMC); // D0
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource15,GPIO_AF_FSMC); // D1
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource0 ,GPIO_AF_FSMC); // D2
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource1 ,GPIO_AF_FSMC); // D3
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource7 ,GPIO_AF_FSMC); // D4
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource8 ,GPIO_AF_FSMC); // D5
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource9 ,GPIO_AF_FSMC); // D6
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource10,GPIO_AF_FSMC); // D7
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource11,GPIO_AF_FSMC); //A16 => RS
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource7 ,GPIO_AF_FSMC); //NE1 => CS
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource4 ,GPIO_AF_FSMC); //NOE => RD
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource5 ,GPIO_AF_FSMC); //NWE => WR
  
 gpio.GPIO_Pin=GPIO_Pin_0|GPIO_Pin_1|GPIO_Pin_4|GPIO_Pin_5|GPIO_Pin_7|GPIO_Pin_11|GPIO_Pin_14|GPIO_Pin_15;
 gpio.GPIO_Mode=GPIO_Mode_AF;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOD,&gpio);

 gpio.GPIO_Pin=GPIO_Pin_7|GPIO_Pin_8|GPIO_Pin_9|GPIO_Pin_10;
 gpio.GPIO_Mode=GPIO_Mode_AF;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOE,&gpio);
 
 //LCD_RESET B11
 gpio.GPIO_Pin=GPIO_Pin_11;
 gpio.GPIO_Mode=GPIO_Mode_OUT;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOB,&gpio);
}

void FSMC_Init(void)
{
 FSMC_NORSRAMTimingInitTypeDef fsmcTiming;
 FSMC_NORSRAMInitTypeDef fsmc;
 
 RCC_AHB3PeriphClockCmd(RCC_AHB3Periph_FSMC,ENABLE);

 fsmcTiming.FSMC_AddressSetupTime=0;
 fsmcTiming.FSMC_AddressHoldTime=0;
 fsmcTiming.FSMC_DataSetupTime=2;
 fsmcTiming.FSMC_BusTurnAroundDuration=0;
 fsmcTiming.FSMC_CLKDivision=0;
 fsmcTiming.FSMC_DataLatency=0;
 fsmcTiming.FSMC_AccessMode=FSMC_AccessMode_A;

 fsmc.FSMC_Bank=FSMC_Bank1_NORSRAM1;
 fsmc.FSMC_DataAddressMux=FSMC_DataAddressMux_Disable;
 fsmc.FSMC_MemoryType=FSMC_MemoryType_SRAM;
 fsmc.FSMC_MemoryDataWidth=FSMC_MemoryDataWidth_8b;
 fsmc.FSMC_BurstAccessMode=FSMC_BurstAccessMode_Disable;
 fsmc.FSMC_WaitSignalPolarity=FSMC_WaitSignalPolarity_Low;
 fsmc.FSMC_WrapMode=FSMC_WrapMode_Disable;
 fsmc.FSMC_WaitSignalActive=FSMC_WaitSignalActive_BeforeWaitState;
 fsmc.FSMC_WriteOperation=FSMC_WriteOperation_Enable;
 fsmc.FSMC_WaitSignal=FSMC_WaitSignal_Disable;
 fsmc.FSMC_AsynchronousWait=FSMC_AsynchronousWait_Disable;
 fsmc.FSMC_ExtendedMode=FSMC_ExtendedMode_Disable;
 fsmc.FSMC_WriteBurst=FSMC_WriteBurst_Disable;
 fsmc.FSMC_ReadWriteTimingStruct=&fsmcTiming;
 fsmc.FSMC_WriteTimingStruct=&fsmcTiming;

 FSMC_NORSRAMInit(&fsmc); 
 FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM1,ENABLE);
}

void Frame_To_LCD(void)
{
 register u32* FB=(u32*)VideoBuffer;
 register u32 i=(W*H)>>1;
 while(i--)OLED_Reg_DAT32=FB++[0];
}

void Frame_To_LCD_DMA_Init(void)
{
 DMA_InitTypeDef DMA_InitStructure;

 RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_DMA2,ENABLE);

 DMA_DeInit(DMA2_Stream0);

 DMA_InitStructure.DMA_Channel = DMA_Channel_0;  
 DMA_InitStructure.DMA_PeripheralBaseAddr=(u32)&VideoBuffer[0];
 DMA_InitStructure.DMA_Memory0BaseAddr   =(u32)&OLED_Reg_DAT32;
 DMA_InitStructure.DMA_DIR               =DMA_DIR_MemoryToMemory;
 DMA_InitStructure.DMA_BufferSize        =(W*H)>>1;
 DMA_InitStructure.DMA_PeripheralInc     =DMA_PeripheralInc_Enable;
 DMA_InitStructure.DMA_MemoryInc         =DMA_MemoryInc_Disable;
 DMA_InitStructure.DMA_PeripheralDataSize=DMA_PeripheralDataSize_Word;
 DMA_InitStructure.DMA_MemoryDataSize    =DMA_MemoryDataSize_Word;
 DMA_InitStructure.DMA_Mode              =DMA_Mode_Normal;
 DMA_InitStructure.DMA_Priority          =DMA_Priority_High; //High
 DMA_InitStructure.DMA_FIFOMode          =DMA_FIFOMode_Disable;
 DMA_InitStructure.DMA_FIFOThreshold     =DMA_FIFOThreshold_Full;
 DMA_InitStructure.DMA_MemoryBurst       =DMA_MemoryBurst_Single;
 DMA_InitStructure.DMA_PeripheralBurst   =DMA_PeripheralBurst_Single;

 DMA_Init(DMA2_Stream0,&DMA_InitStructure);
}

void Frame_To_LCD_DMA(void)
{
 while(DMA_GetCmdStatus(DMA2_Stream0)!=DISABLE); //while(DMA_GetCurrentMemoryTarget(DMA2_Stream0));
 DMA_ClearITPendingBit(DMA2_Stream0,DMA_IT_TCIF0);
 DMA_Cmd(DMA2_Stream0,ENABLE);
}

void DAC_Clock_Config(void)
{
 RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA|RCC_AHB1Periph_DMA1,ENABLE);
 RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6|RCC_APB1Periph_DAC,ENABLE);
}

void DAC_GPIO_Config(void)
{
 GPIO_InitTypeDef GPIO_A;

 GPIO_A.GPIO_Pin  =GPIO_Pin_4;
 GPIO_A.GPIO_Mode =GPIO_Mode_AN;
 GPIO_A.GPIO_Speed=GPIO_Speed_2MHz;
 GPIO_A.GPIO_PuPd =GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOA,&GPIO_A);

 GPIO_A.GPIO_Pin  =GPIO_Pin_0;   //Pause
 GPIO_A.GPIO_Mode =GPIO_Mode_IN;
 GPIO_Init(GPIOA,&GPIO_A);
}

void DAC_TIM_Config(void)
{
 TIM_TimeBaseInitTypeDef TIM6_TimeBase;

 TIM_TimeBaseStructInit(&TIM6_TimeBase);
 TIM6_TimeBase.TIM_Period       =TIM_PERIOD-1;
 TIM6_TimeBase.TIM_Prescaler    =1-1;
 TIM6_TimeBase.TIM_ClockDivision=0;
 TIM6_TimeBase.TIM_CounterMode  =TIM_CounterMode_Up;
 TIM_TimeBaseInit(TIM6,&TIM6_TimeBase);
 TIM_SelectOutputTrigger(TIM6,TIM_TRGOSource_Update);
 TIM_SetCounter(TIM6,0);
}

static void DAC_DMA_IRQ_Config()
{
 DMA_InitTypeDef  DMA_INIT;
 NVIC_InitTypeDef NVIC_InitStructure;

 DMA_DeInit(DMA1_Stream5);

 DMA_INIT.DMA_Channel           =DMA_Channel_7;
 DMA_INIT.DMA_PeripheralBaseAddr=DAC_DHR12R1;
 DMA_INIT.DMA_Memory0BaseAddr   =(u32)&AudioBuffer;
 DMA_INIT.DMA_DIR               =DMA_DIR_MemoryToPeripheral;
 DMA_INIT.DMA_BufferSize        =L<<1;
 DMA_INIT.DMA_PeripheralInc     =DMA_PeripheralInc_Disable;
 DMA_INIT.DMA_MemoryInc         =DMA_MemoryInc_Enable;
 DMA_INIT.DMA_PeripheralDataSize=DMA_PeripheralDataSize_HalfWord;
 DMA_INIT.DMA_MemoryDataSize    =DMA_MemoryDataSize_HalfWord;
 DMA_INIT.DMA_Mode              =DMA_Mode_Circular;
 DMA_INIT.DMA_Priority          =DMA_Priority_VeryHigh; //VeryHigh
 DMA_INIT.DMA_FIFOMode          =DMA_FIFOMode_Disable;
 DMA_INIT.DMA_FIFOThreshold     =DMA_FIFOThreshold_HalfFull;
 DMA_INIT.DMA_MemoryBurst       =DMA_MemoryBurst_Single;
 DMA_INIT.DMA_PeripheralBurst   =DMA_PeripheralBurst_Single;
 DMA_Init(DMA1_Stream5,&DMA_INIT);

 NVIC_InitStructure.NVIC_IRQChannel=DMA1_Stream5_IRQn;
 NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority=0;
 NVIC_InitStructure.NVIC_IRQChannelSubPriority=0;
 NVIC_InitStructure.NVIC_IRQChannelCmd=ENABLE;
 NVIC_Init(&NVIC_InitStructure);
 DMA_ITConfig(DMA1_Stream5,DMA_IT_HT|DMA_IT_TC,ENABLE);
}

static void DAC_Config()
{
 DAC_InitTypeDef  DAC_INIT;

 DAC_INIT.DAC_Trigger       =DAC_Trigger_T6_TRGO;
 DAC_INIT.DAC_WaveGeneration=DAC_WaveGeneration_None;
 DAC_INIT.DAC_OutputBuffer  =DAC_OutputBuffer_Enable;
 DAC_Init(DAC_Channel_1,&DAC_INIT);
}

void DAC_ON(void)
{
 TIM_Cmd(TIM6,ENABLE);
 DMA_Cmd(DMA1_Stream5,ENABLE);
 DAC_Cmd(DAC_Channel_1,ENABLE);
 DAC_DMACmd(DAC_Channel_1,ENABLE);
 __enable_irq();
}

void DAC_OFF(void)
{
 __disable_irq();
 TIM_Cmd(TIM6,DISABLE);
 DMA_Cmd(DMA1_Stream5,DISABLE);
 DAC_Cmd(DAC_Channel_1,DISABLE);
 DAC_DMACmd(DAC_Channel_1,DISABLE);
}

volatile s32 ws=-1;

void DMA1_Stream5_IRQHandler(void)
{
 if(DMA_GetITStatus(DMA1_Stream5,DMA_IT_HTIF5))      //DMA Half Transfer Complete
 {
  DMA_ClearITPendingBit(DMA1_Stream5,DMA_IT_HTIF5);
  ws=0;
 }
 else if(DMA_GetITStatus(DMA1_Stream5,DMA_IT_TCIF5)) //DMA Transfer Complete
 {
  DMA_ClearITPendingBit(DMA1_Stream5,DMA_IT_TCIF5);
  ws=L;
 }
}

#define F_getc(c)              \
{                              \
 register int r=Fgetc(fv);     \
 if(r==EOF)NVIC_SystemReset(); \
 c=r;                          \
}                              \

int main(void)
{
 u8 c[2];

 FSMC_GPIO_Init();
 FSMC_Init();

 DAC_Clock_Config();
 DAC_GPIO_Config();
 DAC_TIM_Config();
 DAC_DMA_IRQ_Config();
 DAC_Config();

 OLED_Reset();
 OLED_Init();

 OLED_Clear(0);
 OLED_ON();

 FatFs_Init();

 devid.jpic=JPEGBuffer;

 OLED_Position(0,W-1,0,H-1);

 Frame_To_LCD_DMA_Init();

 if(SC=='NEXT')SN++; //    Video/Audio
 else
 {
  SC='NEXT';
  SN=0;
 }

 PlayNext:

 sprintf(SS,"%d",SN);
 fv=Fopen(strcat(SS,".avi"),"rb");    //12 FPS video MJPEG

 if(fv==NULL)
 {
  SN=0;
  goto PlayNext;
 }

 sprintf(SS,"%d",SN);
 drflac *pFlac=drflac_open_file(strcat(SS,".fla")); //audio FLAC

 memset((void*)AudioBuffer,0,sizeof(AudioBuffer));

 DAC_ON();

 while(1)
 {

  while(ws==-1);

  drflac_read_s16(pFlac,L,(void*)&AudioBuffer[ws]);
  for(u32 i=ws;i<ws+L;i++)AudioBuffer[i]^=2048;

  ws=-1;

  c[0]=0;

  ScanFFD8:
  while(c[0]!=0xFF)F_getc(c[0]);
  F_getc(c[0]);
  if(c[0]!=0xD8)goto ScanFFD8;

  *(u16*)JPEGBuffer=0xD8FF;

  devid.jsize=2;
  c[0]=0;

  ScanFFD9:
  F_getc(c[1]);
  JPEGBuffer[devid.jsize]=c[1];
  if(*(u16*)c==0xD9FF)goto JPEG;

  c[0]=c[1];
  devid.jsize++;
  goto ScanFFD9;

  JPEG:
  devid.joffset=0;

  jd_prepare(&jdec,in_func,WORK,sizeof(WORK),&devid);
  jd_decomp(&jdec,out_func,0);

  Frame_To_LCD_DMA(); //Frame_To_LCD();

  if(PAUSE) /// 
  {
   DAC_OFF();
   delay_ms(100);
   while(PAUSE);
   delay_ms(100);
   while(!PAUSE);
   delay_ms(100);
   while(PAUSE);
   delay_ms(100);
   DAC_ON();
  }

 }
}
