/*-----------------------------------------------------------------------*/
/* Low level disk I/O module skeleton for FatFs     (C)ChaN, 2016        */
/*-----------------------------------------------------------------------*/
/* If a working storage control module is available, it should be        */
/* attached to the FatFs via a glue function rather than modifying it.   */
/* This is an example of glue functions to attach various exsisting      */
/* storage control modules to the FatFs module with a defined API.       */
/*-----------------------------------------------------------------------*/

#include "diskio.h"             /* FatFs lower layer API */
#include "sdio_sd.h"

/*-----------------------------------------------------------------------*/
/* Get Drive Status                                                      */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
        BYTE pdrv               /* Physical drive nmuber to identify the drive */
)
{
 return 0;
}

/*-----------------------------------------------------------------------*/
/* Inidialize a Drive                                                    */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
        BYTE pdrv                               /* Physical drive nmuber to identify the drive */
)
{
 return 0;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
        BYTE pdrv,              /* Physical drive nmuber to identify the drive */
        BYTE *buff,             /* Data buffer to store read data */
        DWORD sector,   /* Start sector in LBA */
        UINT count              /* Number of sectors to read */
)
{
#if defined (SD_POLLING_MODE)
 for(int i=0;i<count;i++)SD_ReadBlock(buff+(i<<9),(sector+i)<<9,512);
#endif

#if defined (SD_DMA_MODE)
 SD_ReadMultiBlocksFIXED(buff,sector,512,count);
 SD_WaitReadOperation();
 while(SD_GetStatus()!=SD_TRANSFER_OK);
#endif

 return RES_OK;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

DRESULT disk_write (
        BYTE pdrv,                      /* Physical drive nmuber to identify the drive */
        const BYTE *buff,       /* Data to be written */
        DWORD sector,           /* Start sector in LBA */
        UINT count                      /* Number of sectors to write */
)
{

#if 0

#if defined (SD_POLLING_MODE)
 for(int i=0;i<count;i++)SD_WriteBlock((uint8_t*)buff+(i<<9),(sector+i)<<9,512);
#endif

#if defined (SD_DMA_MODE)
 SD_WriteMultiBlocksFIXED((uint8_t*)buff,sector,512,count);
 SD_WaitWriteOperation();
 while(SD_GetStatus()!=SD_TRANSFER_OK);
#endif

#endif

 return RES_OK;  
}

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

DRESULT disk_ioctl (
        BYTE pdrv,              /* Physical drive nmuber (0..) */
        BYTE cmd,               /* Control code */
        void *buff              /* Buffer to send/receive control data */
)
{
 return RES_OK;
}
