#include "STM32F4.h"

#include "avcodec.h"
#include "ima.h"

#include "sdio_sd.h"
#include "PC_FAT.h"

#include "memory.h"

#include "OLED.inc"

#include "colortab.inc"
#include "rgb_2_pix.inc"

#define W 160
#define H 128

#define CPU_FREQ 168000000
#define F 24000 
#define FPS 12

#define I 128

#define TIM_PERIOD ((u32)((CPU_FREQ/8)/F))

#define VSYNC ((u32)(CPU_FREQ/FPS))

#define DAC_DHR12L1 0x4000740C

#define DWT_CYCCNT  *(volatile u32*)0xE0001004
#define DWT_CONTROL *(volatile u32*)0xE0001000
#define SCB_DEMCR   *(volatile u32*)0xE000EDFC

#define SC (*(u32*)0x1000FFF8)
#define SN (*(u32*)0x1000FFFC)

#define PAUSE GPIO_ReadInputDataBit(GPIOA,GPIO_Pin_0)

#define DMA_IT_ALL5 (DMA_IT_FEIF5|DMA_IT_DMEIF5|DMA_IT_TEIF5|DMA_IT_HTIF5|DMA_IT_TCIF5)

#define CLI       \
{                 \
 __disable_irq(); \
 __DSB();         \
 __ISB();         \
}                 \

#define STI __enable_irq();

#define F_getc(c)        \
{                        \
 CLI                     \
 f_read(inpf,&c,1,NULL); \
 STI                     \
}                        \

FIL fv,fa;

char SS[10];

AVCodecContext c;
AVFrame picture;
int got_picture;

u8  *MPEGBuffer;
s16 *AudioBuffer;

int *u_b_tab,*u_g_tab,*v_g_tab,*v_r_tab;
unsigned int *r_2_pix,*g_2_pix,*b_2_pix;

void CreateYUVTab_16()
{
	u_b_tab = &((int*)colortab_LUT)[0*256];
	u_g_tab = &((int*)colortab_LUT)[1*256];
	v_g_tab = &((int*)colortab_LUT)[2*256];
	v_r_tab = &((int*)colortab_LUT)[3*256];

	r_2_pix = (&((unsigned int*)rgb_2_pix_LUT)[0*768])+256;
	g_2_pix = (&((unsigned int*)rgb_2_pix_LUT)[1*768])+256;
	b_2_pix = (&((unsigned int*)rgb_2_pix_LUT)[2*768])+256;
}

void DisplayYUV_16(unsigned char *y, unsigned char *u, unsigned char *v, int width, int height, int src_ystride, int src_uvstride, int dst_ystride)
{
	int i, j, k;
	int r, g, b , rgb;

	int yy, ub, ug, vg, vr;

	unsigned char* yoff;
	unsigned char* uoff;
	unsigned char* voff;
	
	int width2 = width/2;
	int height2 = height/2;
	
	if(width2>W/2)
	{
		width2=W/2;

		y+=(width-W)/4*2;
		u+=(width-W)/4;
		v+=(width-W)/4;
	}

	if(height2>H)height2=H;

/*
  2x2:

014589CD
2367ABEF...
*/
	for(j=0; j<height2; j++)
	{
		yoff = y + j * 2 * src_ystride;
		uoff = u + j * src_uvstride;
		voff = v + j * src_uvstride;

		for(k=0;k<=src_ystride;k+=src_ystride)for(i=0;i<width2;i++)
		{
			ub = u_b_tab[*(uoff+i)];
			ug = u_g_tab[*(uoff+i)];
			vg = v_g_tab[*(voff+i)];
			vr = v_r_tab[*(voff+i)];

			yy=*(yoff+k+(i<<1));

			b=yy+ub;
			g=yy-ug-vg;
			r=yy+vr;

			rgb=r_2_pix[r]|g_2_pix[g]|b_2_pix[b];

			yy=*(yoff+k+((i<<1)|1));

			b=yy+ub;
			g=yy-ug-vg;
			r=yy+vr;

			__asm
                        {
			 PKHBT rgb,rgb,r_2_pix[r]|g_2_pix[g]|b_2_pix[b],LSL 16
			 REV16 OLED_Reg_DAT32,rgb
			}

		}
  }
}

int FindStartCode(unsigned char *Buf, int zeros_in_startcode)
{
  int info;
  int i;

  info = 1;
  for (i = 0; i < zeros_in_startcode; i++)
    if(Buf[i] != 0)
      info = 0;

  if(Buf[i] != 1)
    info = 0;
  return info;
}

int getNextNal(FIL* inpf,  unsigned char* Buf)
{
	int pos = 0;
	int StartCodeFound = 0;
	int info2 = 0;
	int info3 = 0;

while(!f_eof(inpf))
{
 F_getc(Buf[pos]);
 pos++;
 if(Buf[pos-1]!=0)goto OK;
}

NVIC_SystemReset();

OK:

	while(!StartCodeFound)
	{
		if(f_eof(inpf))NVIC_SystemReset();

		F_getc(Buf[pos++]);

		info3 = FindStartCode(&Buf[pos-4], 3);

		if(info3 != 1)info2 = FindStartCode(&Buf[pos-3], 2);

		StartCodeFound = (info2 == 1 || info3 == 1);
	}

        CLI
	Fseek(inpf, -4, SEEK_CUR);
        STI

	return pos - 4;
}

void FSMC_GPIO_Init(void)
{
 GPIO_InitTypeDef gpio;
 
 RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOB|RCC_AHB1Periph_GPIOD|RCC_AHB1Periph_GPIOE,ENABLE);
 
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource14,GPIO_AF_FSMC); // D0
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource15,GPIO_AF_FSMC); // D1
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource0 ,GPIO_AF_FSMC); // D2
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource1 ,GPIO_AF_FSMC); // D3
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource7 ,GPIO_AF_FSMC); // D4
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource8 ,GPIO_AF_FSMC); // D5
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource9 ,GPIO_AF_FSMC); // D6
 GPIO_PinAFConfig(GPIOE,GPIO_PinSource10,GPIO_AF_FSMC); // D7
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource11,GPIO_AF_FSMC); //A16 => RS
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource7 ,GPIO_AF_FSMC); //NE1 => CS
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource4 ,GPIO_AF_FSMC); //NOE => RD
 GPIO_PinAFConfig(GPIOD,GPIO_PinSource5 ,GPIO_AF_FSMC); //NWE => WR
  
 gpio.GPIO_Pin=GPIO_Pin_0|GPIO_Pin_1|GPIO_Pin_4|GPIO_Pin_5|GPIO_Pin_7|GPIO_Pin_11|GPIO_Pin_14|GPIO_Pin_15;
 gpio.GPIO_Mode=GPIO_Mode_AF;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOD,&gpio);

 gpio.GPIO_Pin=GPIO_Pin_7|GPIO_Pin_8|GPIO_Pin_9|GPIO_Pin_10;
 gpio.GPIO_Mode=GPIO_Mode_AF;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOE,&gpio);
 
 //LCD_RESET B11
 gpio.GPIO_Pin=GPIO_Pin_11;
 gpio.GPIO_Mode=GPIO_Mode_OUT;
 gpio.GPIO_Speed=GPIO_Speed_100MHz;
 gpio.GPIO_OType=GPIO_OType_PP;
 gpio.GPIO_PuPd=GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOB,&gpio);
}

void FSMC_Init(void)
{
 FSMC_NORSRAMTimingInitTypeDef fsmcTiming;
 FSMC_NORSRAMInitTypeDef fsmc;
 
 RCC_AHB3PeriphClockCmd(RCC_AHB3Periph_FSMC,ENABLE);

 fsmcTiming.FSMC_AddressSetupTime=0;
 fsmcTiming.FSMC_AddressHoldTime=0;
 fsmcTiming.FSMC_DataSetupTime=2;
 fsmcTiming.FSMC_BusTurnAroundDuration=0;
 fsmcTiming.FSMC_CLKDivision=0;
 fsmcTiming.FSMC_DataLatency=0;
 fsmcTiming.FSMC_AccessMode=FSMC_AccessMode_A;

 fsmc.FSMC_Bank=FSMC_Bank1_NORSRAM1;
 fsmc.FSMC_DataAddressMux=FSMC_DataAddressMux_Disable;
 fsmc.FSMC_MemoryType=FSMC_MemoryType_SRAM;
 fsmc.FSMC_MemoryDataWidth=FSMC_MemoryDataWidth_8b;
 fsmc.FSMC_BurstAccessMode=FSMC_BurstAccessMode_Disable;
 fsmc.FSMC_WaitSignalPolarity=FSMC_WaitSignalPolarity_Low;
 fsmc.FSMC_WrapMode=FSMC_WrapMode_Disable;
 fsmc.FSMC_WaitSignalActive=FSMC_WaitSignalActive_BeforeWaitState;
 fsmc.FSMC_WriteOperation=FSMC_WriteOperation_Enable;
 fsmc.FSMC_WaitSignal=FSMC_WaitSignal_Disable;
 fsmc.FSMC_AsynchronousWait=FSMC_AsynchronousWait_Disable;
 fsmc.FSMC_ExtendedMode=FSMC_ExtendedMode_Disable;
 fsmc.FSMC_WriteBurst=FSMC_WriteBurst_Disable;
 fsmc.FSMC_ReadWriteTimingStruct=&fsmcTiming;
 fsmc.FSMC_WriteTimingStruct=&fsmcTiming;

 FSMC_NORSRAMInit(&fsmc); 
 FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM1,ENABLE);
}

void DAC_Clock_Config(void)
{
 RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA|RCC_AHB1Periph_DMA1,ENABLE);
 RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM6|RCC_APB1Periph_DAC,ENABLE);
}

void DAC_GPIO_Config(void)
{
 GPIO_InitTypeDef GPIO_A;

 GPIO_A.GPIO_Pin  =GPIO_Pin_4;
 GPIO_A.GPIO_Mode =GPIO_Mode_AN;
 GPIO_A.GPIO_Speed=GPIO_Speed_2MHz;
 GPIO_A.GPIO_PuPd =GPIO_PuPd_NOPULL;
 GPIO_Init(GPIOA,&GPIO_A);

 GPIO_A.GPIO_Pin  =GPIO_Pin_0;   //Pause
 GPIO_A.GPIO_Mode =GPIO_Mode_IN;
 GPIO_Init(GPIOA,&GPIO_A);
}

void DAC_TIM_Config(void)
{
 TIM_TimeBaseInitTypeDef TIM6_TimeBase;

 TIM_TimeBaseStructInit(&TIM6_TimeBase);

 TIM6_TimeBase.TIM_Period       =TIM_PERIOD-1;
 TIM6_TimeBase.TIM_Prescaler    =1-1;
 TIM6_TimeBase.TIM_ClockDivision=TIM_CKD_DIV1;
 TIM6_TimeBase.TIM_CounterMode  =TIM_CounterMode_Up;
 TIM_TimeBaseInit(TIM6,&TIM6_TimeBase);
 TIM_SelectOutputTrigger(TIM6,TIM_TRGOSource_Update);
 TIM_SetCounter(TIM6,0);
}

static void DAC_DMA_IRQ_Config()
{
 DAC_InitTypeDef  DAC_INIT;
 DMA_InitTypeDef  DMA_INIT;
 NVIC_InitTypeDef NVIC_InitStructure;

 DAC_INIT.DAC_Trigger       =DAC_Trigger_T6_TRGO;
 DAC_INIT.DAC_WaveGeneration=DAC_WaveGeneration_None;
 DAC_INIT.DAC_OutputBuffer  =DAC_OutputBuffer_Enable;
 DAC_Init(DAC_Channel_1,&DAC_INIT);

 DMA_DeInit(DMA1_Stream5);

 DMA_INIT.DMA_Channel           =DMA_Channel_7;
 DMA_INIT.DMA_PeripheralBaseAddr=DAC_DHR12L1;
 DMA_INIT.DMA_Memory0BaseAddr   =(u32)AudioBuffer;
 DMA_INIT.DMA_DIR               =DMA_DIR_MemoryToPeripheral;
 DMA_INIT.DMA_BufferSize        =I<<2;
 DMA_INIT.DMA_PeripheralInc     =DMA_PeripheralInc_Disable;
 DMA_INIT.DMA_MemoryInc         =DMA_MemoryInc_Enable;
 DMA_INIT.DMA_PeripheralDataSize=DMA_PeripheralDataSize_HalfWord;
 DMA_INIT.DMA_MemoryDataSize    =DMA_MemoryDataSize_HalfWord;
 DMA_INIT.DMA_Mode              =DMA_Mode_Circular;
 DMA_INIT.DMA_Priority          =DMA_Priority_VeryHigh; //VeryHigh
 DMA_INIT.DMA_FIFOMode          =DMA_FIFOMode_Disable;
 DMA_INIT.DMA_FIFOThreshold     =DMA_FIFOThreshold_HalfFull;
 DMA_INIT.DMA_MemoryBurst       =DMA_MemoryBurst_Single;
 DMA_INIT.DMA_PeripheralBurst   =DMA_PeripheralBurst_Single;
 DMA_Init(DMA1_Stream5,&DMA_INIT);

 NVIC_InitStructure.NVIC_IRQChannel=DMA1_Stream5_IRQn;
 NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority=0;
 NVIC_InitStructure.NVIC_IRQChannelSubPriority=0;
 NVIC_InitStructure.NVIC_IRQChannelCmd=ENABLE;
 NVIC_Init(&NVIC_InitStructure);
 DMA_ITConfig(DMA1_Stream5,DMA_IT_HT|DMA_IT_TC,ENABLE);
}

void DAC_ON(void)
{
 TIM_Cmd(TIM6,ENABLE);
 DMA_Cmd(DMA1_Stream5,ENABLE);
 DAC_Cmd(DAC_Channel_1,ENABLE);
 DAC_DMACmd(DAC_Channel_1,ENABLE);
 STI
}

void DAC_OFF(void)
{
 CLI
 TIM_Cmd(TIM6,DISABLE);
 DMA_Cmd(DMA1_Stream5,DISABLE);
 DAC_Cmd(DAC_Channel_1,DISABLE);
 DAC_DMACmd(DAC_Channel_1,DISABLE);
}

ima_state IMA_State;
u8 IMABuffer[I];

void UpdateAudio(u32 h)
{
 CLI
 f_read(&fa,IMABuffer,I,NULL);
 STI
 if(f_eof(&fa))NVIC_SystemReset();
 ima_decode_mono(&IMA_State,&AudioBuffer[h],IMABuffer,I);
}

void DMA1_Stream5_IRQHandler(void)
{
 if(DMA_GetITStatus(DMA1_Stream5,DMA_IT_HTIF5))UpdateAudio(0);
 if(DMA_GetITStatus(DMA1_Stream5,DMA_IT_TCIF5))UpdateAudio(I<<1);
 DMA_ClearITPendingBit(DMA1_Stream5,DMA_IT_ALL5);
 __DSB();
}

int main(void)
{
 MPEGBuffer=malloc(8*1024); //MPEG Buffer

 AudioBuffer=malloc(I<<3);
 memset(AudioBuffer,0,I<<3);

 FSMC_GPIO_Init();
 FSMC_Init();

 DAC_Clock_Config();
 DAC_GPIO_Config();
 DAC_TIM_Config();
 DAC_DMA_IRQ_Config();

 BKPSRAM_Init();

 OLED_Reset();
 OLED_Init();
 OLED_ON();

 OLED_Clear(0);
 OLED_Position(0,W-1,0,H-1);

 FatFs_Init();

 if(SC=='NEXT')SN++; //    Video/Audio
 else
 {
  SC='NEXT';
  SN=0;
 }

 PlayNext:

 sprintf(SS,"%d",SN);
 if(f_open(&fv,strcat(SS,".264"),FA_READ)) //12 FPS video H264
 {
  SN=0;
  goto PlayNext;
 }

 sprintf(SS,"%d",SN);
 f_open(&fa,strcat(SS,".ima"),FA_READ);    //96 kbps audio IMA ADPCM

 CreateYUVTab_16();

 avcodec_open(&c);

 DAC_ON();

 SCB_DEMCR|=CoreDebug_DEMCR_TRCENA_Msk; //  DWT
 DWT_CONTROL|=DWT_CTRL_CYCCNTENA_Msk;   // 
 DWT_CYCCNT=0;                          // 

 Loop:

 decode_frame(&c,&picture,&got_picture,MPEGBuffer,getNextNal(&fv,MPEGBuffer));

 if(got_picture)
 {
  DisplayYUV_16(picture.data[0],picture.data[1],picture.data[2],c.width,c.height,picture.linesize[0],picture.linesize[1],W);

  while(DWT_CYCCNT<VSYNC); // -
  DWT_CYCCNT-=VSYNC;
 }

 if(PAUSE) /// 
 {
  DAC_OFF();
  DWT_CONTROL&=~DWT_CTRL_CYCCNTENA_Msk;   // 
  delay_ms(100);
  while(PAUSE);
  delay_ms(100);
  while(!PAUSE);
  delay_ms(100);
  while(PAUSE);
  delay_ms(100);
  DAC_ON();
  DWT_CONTROL|=DWT_CTRL_CYCCNTENA_Msk;    // 
 }

 goto Loop;
}
